"""Place objects."""
import numpy as np
from typing import List
import matplotlib.axes


class BoundingBox:
    """
    BoundingBox class to deal with the varying output requirments often needed to
    describe the same geographical box to different APIs.

    Example::
        >>> from sithom.place import BoundingBox
        >>> bbox = BoundingBox([-30, 30], [10, 30], desc="")
        >>> bbox.cartopy() # [lon-, lon+, lat-, lat+]
        [-30, 30, 10, 30]
        >>> bbox.ecmwf() # [lat+, lon-, lat-, lon+]
        [30, -30, 10, 30]
    """

    def __init__(self, lon: List[float], lat: List[float], desc: str = "No Description Given.") -> None:
        """
        Create BBOX.

        Args:
            lon (List[float]): Longitude bounds. Degrees East.
            lat (List[float]): Latitude bounds. Degrees North.
            desc (str): Description of boundary box for debugging. Defaults to "No Description Given.".
        """
        assert len(lon) == 2
        assert len(lat) == 2
        self.lon = lon
        self.lat = lat
        self.desc = desc

    def __repr__(self) -> str:
        """
        Representation string.
        """
        return str(
            [
                ("Latitude bounds", self.lat, "degrees_north"),
                ("Longitude bounds", self.lon, "degrees_east"),
                self.desc,
            ]
        )

    def cartopy(self) -> List[float]:
        """
        Cartopy style bounding box.

        Returns:
            List[float]: [lon-, lon+, lat-, lat+] # [degE, degE, degN, degN]
        """
        return self.lon + self.lat

    def ecmwf(self) -> List[float]:
        """
        ECMWF style bounding box.

        Returns:
            List[float]: [lat+, lon-, lat-, lon+] # [degN, degE, degN, degE]
        """
        return [self.lat[1], self.lon[0], self.lat[0], self.lon[1]]

    def ax_lim(self, ax: matplotlib.axes.Axes) -> None:
        """
        Apply BoundingBox as ax limit to your graph.

        Args:
            ax (matplotlib.axes.Axes): Axes to limit.
        """
        ax.set_xlim(self.lon)
        ax.set_ylim(self.lat)


    def ax_label(self, ax: matplotlib.axes.Axes) -> None:
        """
        Apply BoundingBox as labels to your graph.

        Args:
            ax (matplotlib.axes.Axes): Axes to limit.
        """
        ax.set_xlabel(self.lon)
        ax.set_ylabel(self.lat)

    def pad(self, buffer: float = 1) -> "BoundingBox":
        """
        Pad the BoundingBox by some number of degrees.

        Args:
            buffer (float, optional): How many degrees East and North
                to go out from existing buffer. Defaults to 1.

        Returns:
            BoundingBox: A bounding box that is padded by the buffer.
        """
        return BoundingBox([self.lon[0] - buffer, self.lon[1] + buffer], 
                            [self.lat[0] - buffer, self.lat[1] + buffer], 
                            desc=self.desc + " padded by " + str(buffer) + " degrees")

    def indices_inside(self, lons: np.ndarray, lats: np.ndarray) -> np.ndarray:
        """
        Get indices of points inside the bounding box.

        Currently only works for 1D arrays.

        Args:
            lons (np.ndarray): Longitudes of points to check.
            lats (np.ndarray): Latitudes of points to check.

        Returns:
            np.ndarray: Indices of points inside the bounding box.
        """
        return np.where(
            (lons >= self.lon[0])
            & (lons <= self.lon[1])
            & (lats >= self.lat[0])
            & (lats <= self.lat[1]))[0]


class Point:
    def __init__(self, lon: float, lat: float, desc: str = "No Description Given.") -> None:
        """
        Initialise point.

        Args:
            lon (float): Longitude. Degrees East.
            lat (float): Latitude. Degrees North.
            desc (str): description. Defaults to "No Description Given.".
        """
        self.lon = lon
        self.lat = lat
        self.desc = desc

    def __repr__(self) -> str:
        """
        Representation string.
        """
        return str(
            [
                ("Latitude", self.lat, "degrees_north"),
                ("Longitude", self.lon, "degrees_east"),
                self.desc,
            ]
        )

    def bbox(self, buffer: float = 3) -> BoundingBox:
        """
        Get `BoundingBox` by padding around the `Point` by the buffer of some number of degrees.

        Size of the square is 4 * buffer**2.

        Args:
            buffer (float, optional): How many degrees East and North
                to go out from loc. Defaults to 3.

        Returns:
            BoundingBox: A bounding box like [-91.0715, 28.9511, -89.0715, 30.9511].

        Example::
            >>> from sithom.place import Point
            >>> point = Point(20, 30)
            >>> bbox = point.bbox(2)
            >>> bbox.cartopy() # [lon-, lon+, lat-, lat+]
            [18, 22, 28, 32]
            >>> bbox.ecmwf() # [lat+, lon-, lat-, lon+]
            [32, 18, 28, 22]

        """
        return BoundingBox(
            [self.lon - buffer, self.lon + buffer],
            [self.lat - buffer, self.lat + buffer],
            desc="Bounding Box around point: " + self.desc,
        )



"""Polynomial curve fitting."""
from typing import Callable, Tuple, Sequence, Union, Optional, Literal
import numpy as np
from uncertainties import unumpy as unp, ufloat
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt
from sithom.misc import in_notebook
from sithom.plot import CAM_BLUE, BRICK_RED, OX_BLUE
from sithom.unc import tex_uf

Flt = Union[float, ufloat]


def _cubic(x: float, a: Flt, b: Flt, c: Flt, d: Flt) -> Flt:
    """Fit cubic curve to data"""
    return a * (x ** 3) + b * (x ** 2) + c * x + d


def _parab(x: float, a: Flt, b: Flt, c: Flt) -> Flt:
    """Fit parabola to data."""
    return a * (x ** 2) + b * x + c


def _lin(x: float, a: Flt, b: Flt) -> Flt:
    """Fit line to data using curve_fit."""
    return (a * x) + b


def _lin_0(x: float, a: Flt) -> Flt:
    """Fit line through zero data using curve_fit."""
    return a * x


def _return_func(param: unp.uarray, reg_type: str = "lin") -> Callable:
    """
    Return function so that the linear function only has to be referenced once.

    Args:
        param (np.ndarray): the param.
        reg_type (str, optional): Which fit occured. Defaults to "lin".

    Returns:
        Callable: Function.

    Example of returning the linear function::
        >>> from sithom.curve import _return_func
        >>> func = _return_func(unp.uarray([1, 2], [0.5, 1]), reg_type="lin")
    """

    def lin(x: Sequence[Flt]) -> np.array:
        return _lin(np.array(x), param[0], param[1])

    def lin_0(x: Sequence[Flt]) -> np.array:
        return _lin_0(np.array(x), param[0])

    def parab(x: Sequence[Flt]) -> np.array:
        return _parab(np.array(x), param[0], param[1], param[2])

    def cubic(x: Sequence[Flt]) -> np.array:
        return _cubic(np.array(x), param[0], param[1], param[2], param[3])

    func_dict = {"lin": lin, "lin0": lin_0, "parab": parab, "cubic": cubic}

    assert reg_type in func_dict
    return func_dict[reg_type]


def fit(
    x_npa: Sequence[Union[float, int]],
    y_npa: Sequence[Union[float, int]],
    reg_type: Literal["lin_0", "lin", "parab", "cubic"] = "lin",
) -> Tuple[unp.uarray, Callable]:
    """
    Fit a polynomial curve, with an estimate of the uncertainty.

    Args:
        x_npa (Sequence[Union[float, int]]): The x values to fit.
        y_npa (Sequence[Union[float, int]]): The y values to fit.
        reg_type (str, optional): Which regression to do. Defaults to "lin".

    Returns:
        Tuple[unp.uarray, Callable]: Paramaters with uncertainty,
            and function to put input data into.

    Example of usage::
        >>> import numpy as np
        >>> from sithom.curve import fit
        >>> param, func = fit(np.array([0, 1, 2]), np.array([1, 4, 7]))
        >>> assert np.isclose(param[0].n, 3.0)
        >>> assert np.isclose(param[1].n, 1.00)

    """
    func_dict = {"lin": _lin, "lin0": _lin_0, "parab": _parab, "cubic": _cubic}

    assert reg_type in func_dict

    popt, pcov = curve_fit(func_dict[reg_type], x_npa, y_npa)
    perr = np.sqrt(np.diag(pcov))
    param = unp.uarray(popt, perr)

    return param, _return_func(param, reg_type=reg_type)


def _label(param: Sequence[ufloat]) -> str:
    """Label generator for polynomial.

    Args:
        param (Sequence[ufloat]): Polynomial fit to print out.

    Returns:
        str: Output of polynomial (e.g 'y  = ($2\pm1$) x + 1\pm 2' )

    Examples:
        >>> from sithom.curve import _label
        >>> from uncertainties import ufloat
        >>> _label([ufloat(1, 1), ufloat(1, 1), ufloat(1, 1)])
        'y =  + $\\\\left( 1.0 \\\\pm 1.0 \\\\right)$x$^{2}$ + $\\\\left( 1.0 \\\\pm 1.0 \\\\right)$x + $1.0 \\\\pm 1.0$'
        >>> _label([ufloat(1, 1), ufloat(1, 1), ufloat(0, 0)])
        'y =  + $\\\\left( 1.0 \\\\pm 1.0 \\\\right)$x$^{2}$ + $\\\\left( 1.0 \\\\pm 1.0 \\\\right)$x + $0.0 \\\\pm 0$'
    """
    output = "y = "
    for i in range(len(param)):
        if len(param) - 1 == i:
            output += " + " + tex_uf(param[i], bracket=False, force_latex=True)
        elif len(param) - 2 == i:
            output += " + " + tex_uf(param[i], bracket=True, force_latex=True) + "x"
        else:
            output += (
                " + "
                + tex_uf(param[i], bracket=True, force_latex=True)
                + "x$^{"
                + str(len(param) - 1)
                + "}$"
            )
    return output


def plot(
    x_values: Sequence[Union[float, int]],
    y_values: Sequence[Union[float, int]],
    reg_type: Literal["lin_0", "lin", "parab", "cubic"] = "lin",
    x_label: str = "x label",
    y_label: str = "y label",
    ext: float = 0.05,
    fig_path: Optional[str] = None,
    ax_format: Optional[Literal["both", "x", "y"]] = "both",
) -> Tuple[unp.uarray, Callable]:
    """
    Plot the polynomial.

    Args:
        x_values (Sequence[Union[float, int]]): The x values to fit.
        y_values (Sequence[Union[float, int]]): The y values to fit.
        reg_type (str, optional): Which regression to do. Defaults to "lin".
        x_label (str, optional): X label for plot. e.g.
            r"$\\Delta \\bar{T}_s$ over tropical pacific (pac) region [$\\Delta$ K]"
        y_label (str): Y labelsfor plot. e.g.
            r"$\\Delta \\bar{T}_s$ over nino3.4 region [$\\Delta$ K]"
        ext: how far in percentage terms to extend beyond data.
        fig_path (Optional[str], optional): Path to stor the figure in.
            Defaults to None.
        ax_format (Literal["both", "x", "y"], optional): which axes to format
            in scientific notation. Defaults to "both".

    Returns:
        Tuple[unp.uarray, Callable]: Paramaters with uncertainty,
            function to put data into.

    Example::
        >>> from sithom.curve import plot
        >>> param, func = plot(
        ...                    [-0.1, 0.5, 1.0, 1.5, 2.3, 2.9, 3.5], 
        ...                    [-0.7, 0.1, 0.3, 1.1, 1.5, 2.3, 2.2]
        ...                   )
        >>> "({:.3f}".format(param[0].n) + ", {:.3f})".format(param[0].s)
        '(0.842, 0.078)'
        >>> "({:.3f}".format(param[1].n) + ", {:.3f})".format(param[1].s)
        '(-0.424, 0.161)'
    """
    param, func = fit(x_values, y_values, reg_type=reg_type)
    min_x_data = min(x_values)
    max_x_data = max(x_values)
    min_x_pred = min_x_data - (max_x_data - min_x_data) * ext
    max_x_pred = max_x_data + (max_x_data - min_x_data) * ext
    x_pred = np.linspace(min_x_pred, max_x_pred, num=50)
    y_pred = func(x_pred)
    y_pred_n = unp.nominal_values(y_pred)
    y_pred_s = unp.std_devs(y_pred)
    if len(param) == 1:
        param = list(param)
        param.append(ufloat(0, 0))
    label = _label(param)
    plt.fill_between(
        x_pred, y_pred_n + y_pred_s, y_pred_n - y_pred_s, alpha=0.5, color=CAM_BLUE
    )
    plt.plot(x_pred, y_pred_n, label=label, color=BRICK_RED, alpha=0.7)
    plt.scatter(x_values, y_values, color=OX_BLUE, alpha=0.7)

    if ax_format is not None:
        plt.gca().ticklabel_format(
            axis=ax_format, style="sci", scilimits=(0, 0), useMathText=True
        )

    if len(param) >= 3:
        plt.legend(
            bbox_to_anchor=(-0.15, 1.02, 1.15, 0.102), loc="lower left", mode="expand",
        )
    else:
        plt.legend()

    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.xlim(min_x_pred, max_x_pred)
    plt.tight_layout()

    if fig_path is not None:
        plt.savefig(fig_path)
    if not in_notebook:
        plt.clf()

    return param, func



"""Uncertainties Utilities Module."""
import numpy as np
import matplotlib
from uncertainties import ufloat


def tex_uf(
    ufloat_input: ufloat,
    bracket: bool = False,
    force_latex: bool = False,
    exponential: bool = True,
) -> str:
    """
    A function to take an uncertainties.ufloat, and return a tex containing string
    for plotting, which has the right number of decimal places.

    Args:
        ufloat_input (ufloat): The uncertainties ufloat object.
        bracket (bool, optional): Whether or not to add latex brackets around
            the parameter. Defaults to False.
        force_latex (bool, optional): Whether to force latex output.
            Defaults to False. If false will check matplotlib.rcParams first.
        exponential (bool, optional): Whether to put in scientific notation.
            Defaults to True.

    Returns:
        str: String ready to be added to a graph label.

    Example usage::
        >>> from uncertainties import ufloat
        >>> from sithom.unc import tex_uf
        >>> uf = ufloat(1, 0.5)
        >>> tex_uf(uf, bracket=True, force_latex=True)
            '$\\\\left( 1.0 \\\\pm 0.5 \\\\right)$'
        >>> uf = ufloat(10, 5)
        >>> tex_uf(uf, bracket=True, force_latex=True)
            '$\\\\left( \\\\left(1.0 \\\\pm 0.5\\\\right) \\\\times 10^{1} \\\\right)$'
        >>> tex_uf(ufloat(0.0, 0.0), bracket=True, force_latex=True) # works weirdly
            '$\\\\left( 0.0 \\\\pm 0 \\\\right)$'
        >>> tex_uf(ufloat(2, 0.06), bracket=True, force_latex=True)
            '$\\\\left( 2.0 \\\\pm 0.1 \\\\right)$'

    (Had to add twice as many backslashes for pytest to run.)

    TODO: Fix connsistency to get bullet points right.
    """
    # trying to deal with divide by zero possibility.
    if ufloat_input.n == 0.0:
        exponential_str = ""
    else:
        if exponential and round(np.log10(abs(ufloat_input.n))) != 0:
            exponential_str = "e"
        else:
            exponential_str = ""

    if ufloat_input.n == 0.0 or ufloat_input.s == 0.0:
        decimal_point = 0
    else:
        # ensure that 1 s.f for error means both agree on decimal points.
        decimal_point = round(
            np.log10(abs(ufloat_input.n)) - np.log10(abs(ufloat_input.s))
        )

        # enforce rule that if it starts with a 1, add an extra decimal point.

        if str(ufloat_input.n)[0] == "1":
            if exponential_str == "e":
                decimal_point += 1
            else:
                decimal_point += 2  # weird behaviour

    # check if Latex is engaged
    if matplotlib.rcParams["text.usetex"] is True or force_latex:
        if bracket:
            format_string = (
                "$\\left( {:." + str(decimal_point) + exponential_str + "L} \\right)$"
            )
        else:
            format_string = "${:." + str(decimal_point) + exponential_str + "L}$"
    else:
        if bracket:
            format_string = "({:." + str(decimal_point) + exponential_str + "P})"
        else:
            format_string = "{:." + str(decimal_point) + exponential_str + "P}"
    return format_string.format(ufloat_input)



"""Miscillanious Utilities Module."""
import inspect
from sys import getsizeof


def in_notebook() -> bool:
    """
    Check if in jupyter notebook.

    Taken from this answer:
    https://stackoverflow.com/a/22424821

    Returns:
        bool: whether in jupyter notebook.

    Example of triggering from python terminal::
        >>> from sithom.misc import in_notebook
        >>> in_notebook()
        False
    """
    try:
        # pylint: disable=import-outside-toplevel
        from IPython import get_ipython

        if "IPKernelApp" not in get_ipython().config:  # pragma: no cover
            return False
    except ImportError:
        return False
    except AttributeError:
        return False
    return True


def human_readable_size(num: int, suffix: str = "B") -> str:
    """Convert a number of bytes into human readable format.

    This function is meant as a helper function for `get_byte_size`.

    Args:
        num (int): The number of bytes to convert
        suffix (str, optional): The suffix to use for bytes. Defaults to 'B'.

    Returns:
        str: A human readable version of the number of bytes.

    Example of using human readable sizes::
        >>> from sithom.misc import human_readable_size
        >>> assert human_readable_size(int(10e5)) == "977 KB"
        >>> assert human_readable_size(int(10e13)) == "91 TB"
    """
    assert num >= 0, "Size cannot be negative."
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if num < 1024:
            return f"{num:.0f} {unit}{suffix}"
        num /= 1024.0
    return f"{num:1f} Y{suffix}"


def calculate_byte_size_recursively(obj: object, seen: set = None) -> int:
    """Recursively calculate size of objects in memory in bytes.

    From: https://github.com/bosswissam/pysize.
    Meant as a helper function for `get_byte_size`.

    Args:
        obj (object): The python object to get the size of
        seen (set, optional): This variable is needed to for the recusrive
            function evaluations, to ensure each object only gets counted once.
            Leave it at "None" to get the full byte size of an object. Defaults to None.

    Returns:
        int: The size of the object in bytes.
    """

    # Note: getsizeof alone is not enough, as it only returns the size of the top
    #  level object, not of its member variables/objects. Hence the recursive calls.
    size = getsizeof(obj)
    if seen is None:
        # At first iteration (top level object), initialize 'seen' as empty set
        seen = set()

    obj_id = id(obj)
    if obj_id in seen:
        # If object was already counted, return 0 size to avoid double counting.
        return 0

    # Important: Mark as seen *before* entering recursion to handle
    # self-referential objects
    seen.add(obj_id)

    if hasattr(obj, "__dict__"):
        # handles class objects
        for cls in obj.__class__.__mro__:
            if "__dict__" in cls.__dict__:
                dictionary = cls.__dict__["__dict__"]
                if inspect.isgetsetdescriptor(dictionary) or inspect.ismemberdescriptor(
                    dictionary
                ):
                    # Recursively calculate size of member objects & variables
                    size += calculate_byte_size_recursively(obj.__dict__, seen)
                break

    if isinstance(obj, dict):
        # handles dictionaries
        size += sum((calculate_byte_size_recursively(v, seen) for v in obj.values()))
        size += sum((calculate_byte_size_recursively(k, seen) for k in obj.keys()))
    elif hasattr(obj, "__iter__") and not isinstance(obj, (str, bytes, bytearray)):
        # handles array like objects (need to exclude str, bytes bytearray since they
        #  also implement __iter__)
        size += sum((calculate_byte_size_recursively(i, seen) for i in obj))

    if hasattr(obj, "__slots__"):  # can have __slots__ with __dict__
        size += sum(
            calculate_byte_size_recursively(getattr(obj, s), seen)
            for s in obj.__slots__
            if hasattr(obj, s)
        )
    return size


def get_byte_size(obj: object) -> str:
    """Return human readable size of a python object in bytes.

    Args:
        obj (object): The python object to analyse

    Returns:
        str: Human readable string with the size of the object

    Examples of using for numpy array::
        >>> import numpy as np
        >>> from sithom.misc import get_byte_size
        >>> assert isinstance(get_byte_size(np.zeros(int(10e4))), str)
    """
    return human_readable_size(calculate_byte_size_recursively(obj))



"""Plotting Utilities Module.

Contains generic plotting functions that are used to achieve
consistent and easy to produce plots across the project.

Example:
    Usage with simple plots::

        from sithom.plot import (
            plot_defaults,
            label_subplots,
            get_dim,
            set_dim,
            PALETTE,
            STD_CLR_LIST,
            CAM_BLUE,
            BRICK_RED,
            OX_BLUE,
        )

        plot_defaults(use_tex=True)

        # ---- example set of graphs ---

        import numpy as np
        import matplotlib.pyplot as plt

        fig, axs = plt.subplots(2, 2)

        x = np.linspace(0, np.pi, num=100)
        axs[0, 0].plot(x, np.sin(x), color=STD_CLR_LIST[0])
        axs[0, 1].plot(x, np.cos(x), color=STD_CLR_LIST[1])
        axs[1, 0].plot(x, np.sinc(x), color=STD_CLR_LIST[2])
        axs[1, 1].plot(x, np.abs(x), color=STD_CLR_LIST[3])

        # set size
        set_dim(fig, fraction_of_line_width=1, ratio=(5 ** 0.5 - 1) / 2)

        # label subplots
        label_subplots(axs, start_from=0, fontsize=10)

"""
from typing import Sequence, Tuple, Optional, Literal, List
import itertools
from shutil import which
import numpy as np
import numpy.ma as ma
import pandas as pd
import xarray as xr
import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
from jupyterthemes import jtplot
import cmocean
from sithom.misc import in_notebook


REPORT_WIDTH: float = 398.3386  # in pixels
# Constants from SVM
# Standard color list
STD_CLR_LIST = [
    "#4d2923ff",
    "#494f1fff",
    "#38734bff",
    "#498489ff",
    "#8481baff",
    "#c286b2ff",
    "#d7a4a3ff",
]
_paper_colors = sns.color_palette(STD_CLR_LIST)
# Note: To inspect colors, call `sns.palplot(_paper_colors)`
PALETTE = itertools.cycle(_paper_colors)
CAM_BLUE = "#a3c1ad"
OX_BLUE = "#002147"
BRICK_RED = "#CB4154"


def plot_defaults(use_tex: Optional[bool] = None, dpi: Optional[int] = None) -> None:
    """
    Apply plotting style to produce nice looking figures.

    Call this at the start of a script which uses `matplotlib`.
    Can enable `matplotlib` LaTeX backend if it is available.

    Uses serif font to fit into latex report.

    Args:
        use_tex (bool, optional): Whether or not to use latex matplotlib backend.
            Defaults to False.
        dpi (int, optional): Which dpi to set for the figures.
            Defaults to 600 dpi (high quality) in terminal or 150 dpi for notebooks.
            Larger dpi may be needed for presentations.

    Examples:
        Basic setting for the plotting defaults::

            >>> from sithom.plot import plot_defaults
            >>> plot_defaults()

    """
    # mac needs a different plotting backend...
    # if platform == "darwin":
    #    matplotlib.use("TkAgg")

    if in_notebook():
        jtplot.style(theme="grade3", context="notebook", ticks=True, grid=False)
        if use_tex is None:
            use_tex: bool = False  # assume tex does not exist.
        if dpi is None:
            dpi: int = 150
    else:
        if use_tex is None:
            use_tex: bool = False  # assume tex does not exist.
        if dpi is None:
            dpi = 600  # high quality dpi

    p_general = {
        "font.family": "STIXGeneral",  # Nice serif font, similar to latex default.
        # "font.family": "serif",
        # "font.serif": [],
        # Use 10pt font in plots, to match 10pt font in document
        "axes.labelsize": 10,
        "font.size": 10,
        "figure.dpi": dpi,
        "savefig.dpi": dpi,
        # Make the legend/label fonts a little smaller
        "legend.fontsize": 10,
        "xtick.labelsize": 9,
        "ytick.labelsize": 9,
        # Times.
        "date.autoformatter.year": "%Y",
        "date.autoformatter.month": "%Y-%m",
        "date.autoformatter.day": "%Y-%m-%d",
        "date.autoformatter.hour": "%m-%d %H",
        "date.autoformatter.minute": "%Y-%m-%d %H:%M:%S",
        "date.autoformatter.second": "%H:%M:%S",
        "date.autoformatter.microsecond": "%M:%S.%f",
        # Set the font for maths
        "axes.formatter.use_mathtext": True,
        "mathtext.fontset": "cm",
        # "font.sans-serif": ["DejaVu Sans"],  # gets rid of error messages
        # "font.monospace": [],
        "figure.figsize": get_dim(),
        "figure.autolayout": True,  # turn on tight_layout.
        "lines.linewidth": 1.0,
        "scatter.marker": "X",
        "image.cmap": "viridis",
    }
    matplotlib.rcParams.update(p_general)

    # colorblind optimised colormap as default.
    preferred_style = "seaborn-colorblind"
    if preferred_style in plt.style.available:
        matplotlib.style.use(preferred_style)

    if use_tex and which("latex") is not None:
        p_setting = {
            "pgf.texsystem": "pdflatex",
            "text.usetex": True,
            "pgf.preamble": str(
                r"\usepackage[utf8x]{inputenc} \usepackage[T1]{fontenc}"
                + r"\usepackage[separate -uncertainty=true]{siunitx}"
            ),
        }
    else:
        p_setting = {
            "text.usetex": False,
        }
    matplotlib.rcParams.update(p_setting)


# pylint: disable=too-many-arguments
def label_subplots(
    axs: Sequence[matplotlib.axes.Axes],
    labels: Sequence[str] = [chr(ord("`") + z) for z in range(1, 27)],
    start_from: int = 0,
    fontsize: int = 10,
    x_pos: float = 0.02,
    y_pos: float = 0.95,
    override: Optional[Literal["inside", "outside", "default"]] = None,
) -> None:
    """Adds e.g. (a), (b), (c) at the top left of each subplot panel.

    Labelling order achieved through ravelling the input `list` or `np.array`.

    Args:
        axs (Sequence[matplotlib.axes.Axes]): `list` or `np.array` of
            `matplotlib.axes.Axes`.
        labels (Sequence[str]): A sequence of labels for the subplots.
        start_from (int, optional): skips first `start_from` labels. Defaults to 0.
        fontsize (int, optional): Font size for labels. Defaults to 10.
        x_pos (float, optional): Relative x position of labels. Defaults to 0.02.
        y_pos (float, optional): Relative y position of labels. Defaults to 0.95.
        override (Optional[Literal["inside", "outside", "default"]], optional): Choose a
            preset x_pos, y_pos option to overide choices.
            "Outside" is good for busy colormaps. Defaults to None.

    Returns:
        void; alters the `matplotlib.axes.Axes` objects

    Example:
        Here is an example of using this function::

            >>> import matplotlib.pyplot as plt
            >>> from sithom.plot import label_subplots
            >>> fig, axs = plt.subplots(2, 2)
            >>> label_subplots(axs, start_from=0, fontsize=10)
            >>> fig, axs = plt.subplots(2, 2)
            >>> label_subplots(axs, start_from=4, fontsize=10)

    """
    override_d = {"default": "inside", "outside": [-0.12, 1.12], "inside": [0.02, 0.95]}
    if override is not None:
        # allow redirection to keep DRY.
        if not isinstance(override_d[override], list):
            override = override_d[override]
        if override in override_d:
            x_pos = override_d[override][0]
            y_pos = override_d[override][1]

    if isinstance(axs, list):
        axs = np.asarray(axs)
    assert len(axs.ravel()) + start_from <= len(labels)
    subset_labels = []
    for i in range(len(axs.ravel())):
        subset_labels.append(labels[i + start_from])
    for i, label in enumerate(subset_labels):
        axs.ravel()[i].text(
            x_pos,
            y_pos,
            str("(" + label + ")"),
            color="black",
            transform=axs.ravel()[i].transAxes,
            fontsize=fontsize,
            fontweight="bold",
            va="top",
        )


def get_dim(
    width: float = REPORT_WIDTH,
    fraction_of_line_width: float = 1,
    ratio: float = (5 ** 0.5 - 1) / 2,
) -> Tuple[float, float]:
    """Return figure height, width in inches to avoid scaling in latex.

    Default width is `sithom.constants.REPORT_WIDTH`.
    Default ratio is golden ratio, with figure occupying full page width.

    Args:
        width (float, optional): Textwidth of the report to make fontsizes match.
            Defaults to `sithom.constants.REPORT_WIDTH`.
        fraction_of_line_width (float, optional): Fraction of the document width
            which you wish the figure to occupy.  Defaults to 1.
        ratio (float, optional): Fraction of figure width that the figure height
            should be. Defaults to (5 ** 0.5 - 1)/2.

    Returns:
        fig_dim (tuple):
            Dimensions of figure in inches

    Example:
        Here is an example of using this function::

            >>> from sithom.plot import get_dim
            >>> dim_tuple = get_dim(fraction_of_line_width=1, ratio=(5 ** 0.5 - 1) / 2)
            >>> print("({:.2f},".format(dim_tuple[0]), "{:.2f})".format(dim_tuple[1]))
            (5.51, 3.41)

    """

    # Width of figure
    fig_width_pt = width * fraction_of_line_width

    # Convert from pt to inches
    inches_per_pt = 1 / 72.27

    # Figure width in inches
    fig_width_in = fig_width_pt * inches_per_pt
    # Figure height in inches
    fig_height_in = fig_width_in * ratio

    return (fig_width_in, fig_height_in)


def set_dim(
    fig: matplotlib.figure.Figure,
    width: float = REPORT_WIDTH,
    fraction_of_line_width: float = 1,
    ratio: float = (5 ** 0.5 - 1) / 2,
) -> None:
    """Set aesthetic figure dimensions to avoid scaling in latex.

    Default width is `sithom.constants.REPORT_WIDTH`.
    Default ratio is golden ratio, with figure occupying full page width.

    Args:
        fig (matplotlib.figure.Figure): Figure object to resize.
        width (float): Textwidth of the report to make fontsizes match.
            Defaults to `sithom.constants.REPORT_WIDTH`.
        fraction_of_line_width (float, optional): Fraction of the document width
            which you wish the figure to occupy.  Defaults to 1.
        ratio (float, optional): Fraction of figure width that the figure height
            should be. Defaults to (5 ** 0.5 - 1)/2.

    Returns:
        void; alters current figure to have the desired dimensions

    Example:
        Here is an example of using this function::

            >>> import matplotlib.pyplot as plt
            >>> from sithom.plot import set_dim
            >>> fig, ax = plt.subplots(1, 1)
            >>> set_dim(fig, fraction_of_line_width=1, ratio=(5 ** 0.5 - 1) / 2)
 
    """
    fig.set_size_inches(
        get_dim(width=width, fraction_of_line_width=fraction_of_line_width, ratio=ratio)
    )


def cmap(variable_name: str) -> matplotlib.colors.LinearSegmentedColormap:
    """Get cmap from a variable name string.

    Ideally colormaps for variables should be consistent
    throughout the project, and changed in this function.
    The colormaps are set to be green where there are NaN values,
    as this has a high contrast with the colormaps used, and
    should ordinarily represent land, unless something has gone wrong.


    Args:
        variable_name (str): name of variable to give colormap.

    Returns:
        matplotlib.colors.LinearSegmentedColormap: sensible colormap

    Example:
        Usage example for sea surface temperature::

            >>> from sithom.plot import cmap
            >>> cmap_t = cmap("sst")
            >>> cmap_t = cmap("u")
            >>> cmap_t = cmap("ranom")

    """

    # make the function case insensitive
    variable_name = variable_name.lower()

    # collate the variables into a smaller number
    map_d = {
        "rain": "rain",
        "ranom": "tarn",
        "tarn": "tarn",
        "u": "speed",
        "v": "speed",
        "speed": "speed",
        "sst": "sst",
        "salt": "haline",
        "sss": "haline",
        "haline": "haline",
        "delta": "delta",
    }

    # map to cmocean colormaps
    cmap_map_d = {
        # pylint: disable=no-member
        "rain": cmocean.cm.rain,
        "tarn": cmocean.cm.tarn,
        "sst": cmocean.cm.thermal,
        "haline": cmocean.cm.haline,
        "speed": cmocean.cm.speed,
        "delta": cmocean.cm.balance,
    }

    # get cmap_t
    cmap_t = cmap_map_d[map_d[variable_name]]

    # make the map green-ish for nan values
    cmap_t.set_bad(color="#15b01a")

    return cmap_t


def axis_formatter() -> matplotlib.ticker.ScalarFormatter:
    """Returns axis formatter for scientific notation.

        Returns:
            matplotlib.ticker.ScalarFormatter: An object to pass in to a
                matplotlib operation.

    Examples:
        Using with xarray::

            >>> import xarray as xr
            >>> from sithom.plot import axis_formatter
            >>> da = xr.tutorial.open_dataset("air_temperature").air
            >>> quadmesh = da.isel(time=0).plot(cbar_kwargs={"format": axis_formatter()})

    """

    fit_obj = matplotlib.ticker.ScalarFormatter(useMathText=True)
    fit_obj.set_scientific(True)
    fit_obj.set_powerlimits((0, 0))

    return fit_obj


def _balance(vmin: float, vmax: float) -> Tuple[float, float]:
    """Balance vmin, vmax.

    Args:
        vmin (float): Initial colorbar vmin.
        vmax (float): Initial colorbar vmax.

    Returns:
        Tuple[float, float]: balanced colormap.

    Example::
        >>> from sithom.plot import _balance
        >>> _balance(1.4, 2.5)
        (-2.5, 2.5)
        >>> _balance(-1.0, 0.5)
        (-1.0, 1.0)
    """
    assert vmax > vmin
    return np.min([-vmax, vmin]), np.max([-vmin, vmax])


def lim(
    npa: np.ndarray, percentile: float = 5, balance: bool = False
) -> Tuple[float, float]:
    """Return colorbar limits.

    Args:
        npa (np.ndarray): A numpy ndarray with values in, including nans.
        percentile (float, optional): Ignoring nans, use 5th and 95th percentile. Defaults to "5perc".
        balance (bool, optional): Whether to balance limits around zero.
    
    Returns:
        Tuple[float, float]: (vmin, vmax)

    Example with a Gaussian distribution::
        >>> import numpy as np
        >>> from sithom.plot import lim
        >>> samples = np.random.normal(size=(100, 100, 100, 10))
        >>> vmin, vmax = lim(samples)
        >>> print("({:.1f},".format(vmin), "{:.1f})".format(vmax))
        (-1.6, 1.6)
        >>> vmin, vmax = lim(samples, balance=True)
        >>> print("({:.1f},".format(vmin), "{:.1f})".format(vmax))
        (-1.6, 1.6)
    """

    vmin = np.nanpercentile(npa, percentile)
    vmax = np.nanpercentile(npa, 100 - percentile)

    assert vmax > vmin

    if balance:
        _balance(vmin, vmax)

    return (vmin, vmax)


def pairplot(df: pd.DataFrame) -> None:
    """
    Improved seaborn pairplot from:

    https://stackoverflow.com/a/50835066

    Args:
        df (pd.DataFrame): A data frame.
    """

    def corrfunc(x, y, ax=None, **kws) -> None:
        """Plot the correlation coefficient in the
           top left hand corner of a plot.

        A function to use with seaborn's `map_lower` api.
        """
        corr = ma.corrcoef(ma.masked_invalid(x), ma.masked_invalid(y))
        corr_coeff = corr[0, 1]
        ax = ax or plt.gca()
        ax.annotate(f"ρ = {corr_coeff:.2f}", xy=(0.05, 1.0), xycoords=ax.transAxes)

    g = sns.pairplot(df, corner=True)
    g.map_lower(corrfunc)


def feature_grid(
    ds: xr.Dataset,
    fig_var: List[List[str]],
    units: List[List[str]],
    names: List[List[str]],
    vlim: List[List[Tuple[float, float, str]]],
    super_titles: List[str],
    figsize: Tuple[float, float] = (12, 6), # in inches
    label_size: int = 12,
    supertitle_pos: Tuple[float, float] = (0.4, 1.3),
) -> None:
    """Feature grid plot.

    Args:
        ds (xr.Dataset): Input dataset with single timeslice of data on lon/lat grid.
        fig_var (List[List[str]]): Figure variable names.
        units (List[List[str]]): Units of variables.
        names (List[List[str]]): Names of variables to plot.
        vlim (List[List[Tuple[float, float, str]]]): Colorbar limits, and colorbar cmap.
        super_titles (List[str]): The titles for each column.
        figsize (Tuple[float, float], optional): Defaults to (12, 6).
        label_size (int, optional): Defaults to 12.
        supertitle_pos (Tuple[float, float], optional): Relative position for titles. Defaults to (0.4, 1.3).
    """
    shape = np.array(fig_var).shape
    fig, axs = plt.subplots(*shape, sharex=True, sharey=True, figsize=figsize)
    ckwargs = {
        "label": "",
        "format": axis_formatter(),
        "extend": "neither",
        "extendrect": False,
        "extendfrac": 0,
    }
    for i in range(shape[0]):
        for j in range(shape[1]):
            ckwargs = {"label": "", "format": axis_formatter()}
            if vlim[i][j] is None:
                ds[fig_var[i][j]].plot(ax=axs[i, j], cbar_kwargs=ckwargs)
            else:
                ds[fig_var[i][j]].plot(
                    ax=axs[i, j],
                    vmin=vlim[i][j][0],
                    vmax=vlim[i][j][1],
                    cmap=vlim[i][j][2],
                    cbar_kwargs=ckwargs,
                )
            axs[i, j].set_title("")
            if units[i][j] != "":
                axs[i, j].set_title(
                    names[i][j] + "  [" + units[i][j] + "]" + "    ", size=label_size
                )
            else:
                axs[i, j].set_title(names[i][j] + units[i][j], size=label_size)

            axs[i, j].set_xlabel("")
            axs[i, j].set_ylabel("")

    def supertitle(j, title):
        axs[0, j].text(
            *supertitle_pos, title, transform=axs[0, j].transAxes, size=label_size + 5,
        )  # , verticalalignment='center', rotation=270)

    for i, title in enumerate(super_titles):
        supertitle(i, title)

    if "time" in ds:
        print(ds.time.values)
        fig.suptitle(ds.time.values)



"""setup.py allows the installation of the project by pip."""
import os
import sys
from shutil import rmtree
from setuptools import find_packages, setup, Command
from typing import Dict, List

NAME = "sithom"

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

REQUIRED = [
    "matplotlib",
    "seaborn",
    "jupyterthemes",
    "cmocean",
    "numpy",
    "scipy",
    "pandas",
    "xarray",
    "uncertainties",
    "pooch",  # only for test tutorial.
]

here = os.path.abspath(os.path.dirname(__file__))

# Load the package"s _version.py module as a dictionary.
about: Dict = {}
with open(os.path.join(here, NAME, "_version.py")) as f:
    # pylint: disable=exec-used
    exec(f.read(), about)


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options: List = []

    @staticmethod
    def status(s):
        """Print things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        """Publish package to PyPI."""
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(about["__version__"]))
        os.system("git push --tags")

        sys.exit()


setup(
    name=NAME,
    version=about["__version__"],
    author=about["__author__"],
    author_email="sdat2@cam.ac.uk",
    description="General utility scripts",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=str("https://github.com/sdat2/" + NAME),
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    include_package_data=True,
    install_requires=REQUIRED,
    license="MIT",
    # test_suite=str(NAME +".tests.test_all.suite"),
    # setup_requires=["pytest-runner"],
    # package_dir={"": NAME},
    tests_require=["pytest", "pooch"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Typing :: Typed",
    ],
    python_requires=">=3.8",
    cmdclass={"upload": UploadCommand,},
)



"""Time Utilities Module."""
from typing import Callable
import time
from functools import wraps
import signal
from contextlib import contextmanager


class TimeoutException(Exception):
    """The function has timed out, as the time limit has been reached."""


@contextmanager
def time_limit(seconds: int) -> None:
    """Time limit manager.

    Function taken from:

    https://stackoverflow.com/a/601168

    Args:
        seconds (int): how  many seconds to wait until timeout.

    Example:
        Call a function which will take longer than the time limit::

            >>> import time
            >>> from sithom.time import time_limit, TimeoutException
            >>> def long_function_call():
            ...     for t in range(0, 5):
            ...         print("t=", t, "seconds")
            ...         time.sleep(1.1)
            >>> try:
            ...     with time_limit(3):
            ...         long_function_call()
            ...         assert False
            ... except TimeoutException as e:
            ...     print("Timed out!")
            ... except Exception as e:
            ...     print("A different exception\t", e)
            t= 0 seconds
            t= 1 seconds
            t= 2 seconds
            Timed out!

    Seems not to work for windows:
    
    https://github.com/sdat2/sithom/runs/7343392694?check_suite_focus=true

    """

    def _signal_handler(signum, frame):
        raise TimeoutException("Timed out!")

    signal.signal(signal.SIGALRM, _signal_handler)
    signal.alarm(seconds)

    try:
        yield
    finally:
        signal.alarm(0)


def time_stamp() -> str:
    """
    Return the current local time.

    Returns:
        str: Time string format "%Y-%m-%d %H:%M:%S".
    """
    current_time = time.localtime()
    return time.strftime("%Y-%m-%d %H:%M:%S", current_time)


def hr_time(time_in: float) -> str:
    """
    Return human readable time as string.

    I got fed up with converting the number in my head.
    Probably runs very quickly.

    Args:
        time (float): time in seconds

    Returns:
        str: string to print.

    Example:
        120 seconds to human readable string::

            >>> from sithom.time import hr_time
            >>> hr_time(120)
                '02 min 00 s'
    """
    if time_in < 60:
        output_str = "%2.5f s" % time_in
    elif 60 < time_in < 60 * 60:
        output_str = time.strftime("%M min %S s", time.gmtime(time_in))
    elif 60 * 60 < time_in < 24 * 60 * 60:
        output_str = time.strftime("%H hr %M min %S s", time.gmtime(time_in))
    else:
        output_str = "%2.5f s" % time_in
    return output_str


def timeit(method: Callable) -> Callable:
    """`sithom.timeit` is a wrapper for performance analysis.

    It should return the time taken for a function to run. Alters `log_time` `dict`
    if fed in. Add @timeit to the function you want to time. Function needs
    `**kwargs` if you want it to be able to feed in `log_time` `dict`.

    Args:
        method (Callable):  the function that it takes as an input

    Examples:
        Here is an example with the tracking functionality and without::

            from sithom.time import timeit
            @timeit
            def loop(**kwargs):
                total = 0
                for i in range(int(10e2)):
                    for j in range(int(10e2)):
                        total += 1
            tmp_log_d = {}
            loop(log_time=tmp_log_d)
            print(tmp_log_d["loop"])
            loop()

    """

    @wraps(method)
    def timed(*args, **kw):
        time_start = time.perf_counter()
        result = method(*args, **kw)
        time_end = time.perf_counter()
        # time.gmtime()
        if "log_time" in kw:
            name = kw.get("log_name", method.__name__.lower())
            kw["log_time"][name] = time_end - time_start
            print("%r " % method.__name__, hr_time(time_end - time_start), "\n")
        else:
            print("%r " % method.__name__, hr_time(time_end - time_start), "\n")
        return result

    return timed



"""Xarray utilities module."""
from typing import Union, Sequence
import numpy as np
import xarray as xr


def spatial_mean(
    dataarray: xr.DataArray, x_dim: str = "longitude", y_dim: str = "latitude"
) -> xr.DataArray:
    # pylint: disable=anomalous-backslash-in-string
    """
    Average a datarray over "longitude" and "latitude" coordinates.

    Spatially weighted.

    Originally from:
    https://ncar.github.io/PySpark4Climate/tutorials/Oceanic-Ni%C3%B1o-Index/
    (although their version is wrong as it assumes numpy input is degrees)

    https://numpy.org/doc/stable/reference/generated/numpy.cos.html
    https://numpy.org/doc/stable/reference/generated/numpy.radians.html

    The average should behave like:

    .. math::
        :nowrap:

        \\begin{equation}
            \\bar{T}_{\\text{Lat }}=\\frac{1}{\\text{nLon }}
            \\sum_{i=1}^{\\text{nLon}} T_{ \\text{Lon}, \\; i}
        \\end{equation}

        \\begin{equation}
            \\bar{T}_{\\text{ month }}=
            \\frac{
            \\sum_{j=1}^{\\text{nLat} }
            \\cos \\left(\\text{ Lat }_{j} \\right)
            \\bar{T}_{\\text{ Lat }, \\; j}
            }
            {
            \\sum_{j=1}^{\\text{nLat} }
            \\cos \\left( \\text{ Lat }_{j} \\right)
            }
        \\end{equation}

    Args:
        da (xr.DataArray): da to average.
        x_dim: The longitude dimension name. Defaults to "longitude".
        y_dim: The latitude dimension name. Defaults to "latitude".

    Returns:
        xr.DataArray: Avarage of da.

    Example of calculating and plotting mean timeseries of dataarray::

        >>> import xarray as xr
        >>> from sithom.xr import spatial_mean
        >>> da = xr.tutorial.open_dataset("air_temperature").air
        >>> timeseries_mean = spatial_mean(da, x_dim="lon", y_dim="lat")
    
    timeseries_mean.plot.line()
    """
    # make sure the datarray is the right way round.
    dataarray = mon_increase(dataarray, x_dim=x_dim, y_dim=y_dim)

    # Find mean for each latitude
    mean_lat = dataarray.mean(dim=x_dim)

    # Find Weighted mean of those values
    # https://numpy.org/doc/stable/reference/generated/numpy.cos.html
    # https://numpy.org/doc/stable/reference/generated/numpy.radians.html
    num = (np.cos(np.radians(dataarray[y_dim])) * mean_lat).sum(dim=y_dim)
    denom = np.sum(np.cos(np.radians(dataarray[y_dim])))

    # Find spatially weighted mean
    mean = num / denom

    return mean


def _latexify(units: str) -> str:
    """
    Latexify the units.

    Args:
        ins (str): input string.

    Returns:
        str: latexed output string.

    Examples::
        >>> from sithom.xr import _latexify
        >>> _latexify("m s**-2")
        'm s$^{-2}$'
        >>> _latexify("kg m s**-2")
        'kg m s$^{-2}$'
        >>> _latexify("degree_Celsius")
        '$^{\\\\circ}$C'
        >>> _latexify("degK")
        'K'

    """
    output = ""
    for unit_index in units.split(" "):
        if "**" in unit_index:
            unit, index = unit_index.split("**")
            output += unit + "$^{" + index + "}$"
        else:
            output += unit_index
        output += " "
    output = output.strip(" ")
    unit_d = {"degree_Celsius": r"$^{\circ}$C", "degK": "K"}
    for initial_unit in unit_d:
        if initial_unit in units:
            output = output.replace(initial_unit, unit_d[initial_unit])
    return output


def plot_units(
    xr_obj: Union[xr.DataArray, xr.Dataset],
    x_dim: str = "longitude",
    y_dim: str = "latitude",
) -> Union[xr.DataArray, xr.Dataset]:
    """
    Adding good latex units to make the xarray object plottable.

    Xarray uses "long_name" and "units" attributes for plotting.

    Fails softly.

    Args:
        xr_da (Union[xr.DataArray, xr.Dataset]): Initial datarray/dataset
            (potentially with units for axes).
        x_dim (str): Defaults to "longitude".
        y_dim (str): Defaults to "latitude".

    Returns:
        Union[xr.DataArray, xr.Dataset]: Datarray/Dataset with correct
            units/names for plotting. Assuming that you've given the
            correct x_dim and y_dim for the object.

    Examples of using it::
        >>> import xarray as xr
        >>> from sithom.xr import plot_units
        >>> da = plot_units(xr.tutorial.open_dataset("air_temperature").air)
        >>> da.attrs["units"]
        'K'

    """
    if x_dim in xr_obj.coords:
        xr_obj.coords[x_dim].attrs["units"] = r"$^{\circ}$E"
        xr_obj.coords[x_dim].attrs["long_name"] = "Longitude"
    if y_dim in xr_obj.coords:
        xr_obj.coords[y_dim].attrs["units"] = r"$^{\circ}$N"
        xr_obj.coords[y_dim].attrs["long_name"] = "Latitude"

    if isinstance(xr_obj, xr.Dataset):
        for var in xr_obj:
            if "units" in xr_obj[var].attrs:
                xr_obj[var].attrs["units"] = _latexify(xr_obj[var].attrs["units"])
    elif isinstance(xr_obj, xr.DataArray):
        if "units" in xr_obj.attrs:
            xr_obj.attrs["units"] = _latexify(xr_obj.attrs["units"])

    return xr_obj


def mon_increase(
    xr_obj: Union[xr.Dataset, xr.DataArray],
    x_dim: str = "longitude",
    y_dim: str = "latitude",
) -> Union[xr.Dataset, xr.DataArray]:
    """Make sure that an xarray axes has monotonically increasing values.

    Args:
        xr_obj (Union[xr.Dataset, xr.DataArray]): 
        x_dim (str, optional): x dimension name. Defaults to "longitude".
        y_dim (str, optional): y dimension name. Defaults to "latitude".

    Returns:
        Union[xr.Dataset, xr.DataArray]: xarray object.

    Examples::
        >>> import xarray as xr
        >>> da = xr.tutorial.open_dataset("air_temperature").air
        >>> improved_da = mon_increase(da, x_dim="lon", y_dim="lat")

    """

    def positive_monotonic(indices: Sequence) -> bool:
        return all(indices[i] <= indices[i + 1] for i in range(len(indices) - 1))

    def negative_monotonic(indices: Sequence) -> bool:
        return all(indices[i] >= indices[i + 1] for i in range(len(indices) - 1))

    for var in [x_dim, y_dim]:
        if negative_monotonic(xr_obj.coords[var].values):
            xr_obj = xr_obj.reindex(**{var: xr_obj.coords[var][::-1]})
        elif not positive_monotonic(xr_obj.coords[var].values):
            assert False

    return xr_obj
